# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2007 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from gedit import encoding_get_utf8, tab_get_from_document
from gtk import ActionGroup, STOCK_CONVERT, CheckMenuItem, MenuItem, Image, STOCK_INFO, \
		RESPONSE_OK, accelerator_parse, ACCEL_VISIBLE, AccelGroup, STOCK_EDIT, RadioMenuItem
from gobject import timeout_add
from os import getcwd
from os.path import dirname

from Logger import debug, error
from symbols.SymbolController import SymbolController
from snippets.SnippetController import SnippetController
from builder.BuildController import BuildController
from toolbar.LaTeXToolbar import LaTeXToolbar
from Settings import Settings
from QuickstartDialog import QuickstartDialog
from tabular.TableDialog import TableDialog
from graphics.GraphicsDialog import GraphicsDialog
from common import ConditionalSource
from outline import OutlineViewContainer
from Installation import Installation
from util import str_bool
from autocomplete.Popup import Popup
from autocomplete.HintPopup import HintPopup
from ipc.DBusController import DBusController
from document.DocumentHelperManager import DocumentHelperManager
from document.DocumentHelper import LANGUAGE_NONE, LANGUAGE_LATEX, LANGUAGE_BIBTEX
from listings.ListingController import ListingController
from BibTeXEntryDialog import BibTeXEntryDialog


class WindowHelper(object):
	def __init__(self, plugin, window):
		# TODO: put some tasks from here to an "exposed" event handler
		
		self.__window = window
		self.__settings = Settings()
		
		
		self.__insert_widgets()
		
		self.__document_helper_manager = DocumentHelperManager(self)
		self.__document_helper_manager.connect("active-language-changed", lambda *x: self.__helper_changed())
		self.__document_helper_manager.connect("active-helper-changed", lambda *x: self.__helper_changed())
		
		self.current_context_id = None

		self.__save_action = self.__ui_manager.get_action("/MenuBar/FileMenu/FileSaveMenu")
		
		self.__build_controller = BuildController(self)
		
		self.__autocomplete_popup = Popup(self.__window)
		self.__hint_popup = HintPopup(self.__window)
		
		self.__settings.connect("profile-saved", lambda *x: self.__update_profiles())
		self.__settings.connect("profile-updated", lambda *x: self.__update_profiles())
		self.__settings.connect("profile-deleted", lambda *x: self.__update_profiles())
		
		self.__dlg_quickstart = None
		self.__dlg_table = None
		self.__dlg_graphics = None
		
		self.__dbus_controller = DBusController(self)
		self.__listing_controller = ListingController(self)
	
	@property
	def window(self):
		return self.__window
	
	@property
	def autocomplete_popup(self):
		return self.__autocomplete_popup
	
	@property
	def hint_popup(self):
		return self.__hint_popup
	
	@property
	def settings(self):
		return self.__settings
	
	@property
	def window(self):
		return self.__window
	
	def __helper_changed(self):
		helper = self.__document_helper_manager.active_helper
		language = helper.get_language()
		
		# outline view
		if language == LANGUAGE_NONE:
			self.__outline_view_container.set_outline_view(None)
		else:
			self.__outline_view_container.set_outline_view(helper.language_helper.outline_view)
		
		# menu items
		mnu_insert_table = self.__ui_manager.get_widget("/MenuBar/ExtraMenu_1/LatexMenu/InsertTableMenu")
		mnu_insert_graphics = self.__ui_manager.get_widget("/MenuBar/ExtraMenu_1/LatexMenu/InsertGraphicsMenu")
		mnu_insert_listing = self.__ui_manager.get_widget("/MenuBar/ExtraMenu_1/LatexMenu/InsertListingMenu")
		mnu_insert_entry = self.__ui_manager.get_widget("/MenuBar/ExtraMenu_1/LatexMenu/InsertEntryMenu")
		mnu_build = self.__ui_manager.get_widget("/MenuBar/ExtraMenu_1/LatexMenu/BuildMenu")
		
		if language == LANGUAGE_LATEX:
			mnu_insert_table.set_sensitive(True)
			mnu_insert_graphics.set_sensitive(True)
			mnu_insert_listing.set_sensitive(True)
			mnu_insert_entry.set_sensitive(False)
			mnu_build.set_sensitive(True)
		elif language == LANGUAGE_BIBTEX:
			mnu_insert_table.set_sensitive(False)
			mnu_insert_graphics.set_sensitive(False)
			mnu_insert_listing.set_sensitive(False)
			mnu_insert_entry.set_sensitive(True)
			mnu_build.set_sensitive(False)
		else:
			mnu_insert_table.set_sensitive(False)
			mnu_insert_graphics.set_sensitive(False)
			mnu_insert_listing.set_sensitive(False)
			mnu_insert_entry.set_sensitive(False)
			mnu_build.set_sensitive(False)
		
		# TODO: show/hide toolbar
		if self.settings["ToolbarMode"] == "auto":
			#self.__toolbar.set_language(language)
			if language == LANGUAGE_LATEX:
				self.__toolbar.show_all()
			else:
				self.__toolbar.hide()
		
		# select menu item
		self.__listen_to_language_toggled = False
		
		if language == LANGUAGE_NONE:
			self.__item_features_none.set_active(True)
		elif language == LANGUAGE_LATEX:
			self.__item_features_latex.set_active(True)
		elif language == LANGUAGE_BIBTEX:
			self.__item_features_bibtex.set_active(True)
		
		self.__listen_to_language_toggled = True
	
	def deactivate(self):
		self.__remove_widgets()
	
	def __pop_status_message(self):
		status_bar = self.window.get_statusbar()
		if self.current_context_id:
			status_bar.pop(self.current_context_id)
		self.current_context_id = None
		
		return False		# do not call again
	
	def push_status_message(self, message):
		"""
		Push a message to the window's status bar and start a timer that pops it again
		"""
		status_bar = self.window.get_statusbar()
		self.current_context_id = status_bar.get_context_id(message)
		status_bar.push(self.current_context_id, message)
	
		timeout_add(3000, self.__pop_status_message)
	
	@property
	def document_helper_manager(self):
		return self.__document_helper_manager
		
	def save_active_document(self):
		self.__save_action.activate()
	
	def __insert_widgets(self):
		#
		# attach side panels
		#
		
		debug(self, "__insert_widgets")
		
		self.__symbol_controller = SymbolController(self)
		
		self.__outline_view_container = OutlineViewContainer()
		
		self.__snippet_controller = SnippetController(self)
		
		img_outline = Image()
		img_outline.set_from_file(Installation().path + "/pixmaps/outline.png")
		
		img_snippets = Image()
		img_snippets.set_from_file(Installation().path + "/pixmaps/snippet.png")
		
		self.__window.get_side_panel().add_item(self.__outline_view_container, _("Outline"), img_outline)
		
		#
		# attach bottom panels
		#
		
		debug(self, "__insert_widgets", "adding menu entries")
		
		#
		# insert menu items
		#
		self.__ui_manager = self.__window.get_ui_manager()
		self.__action_group = ActionGroup("LaTeXPluginActions")
		self.__action_group.add_actions([
				#("Archive", None, "Archive LaTeX Project...", "<Ctrl><Alt>A", "Create an archive of a LaTeX source file and all files referenced by it", lambda x: self.__archive()),
				
				("Comment", None, "Comment", "<Ctrl><Alt>C", None, lambda x: self.__comment_selection()),
				("Uncomment", None, "Uncomment", "<Ctrl><Alt>U", None, lambda x: self.__comment_selection(uncomment=True)),
				
				("Latex", None, "LaTeX", None, None, None),
				
				("LatexQuickstart", None, "New LaTeX Document...", "<Ctrl><Alt>N", None, lambda x: self.run_quickstart_dialog()),
				("InsertTable", None, "Insert Table...", "<Ctrl><Alt>T", None, lambda x: self.run_table_dialog()),
				("InsertGraphics", None, "Insert Graphics...", "<Ctrl><Alt>G", None, lambda x: self.run_graphics_dialog()),
				# TODO: Ctrl+Alt+L is used for locking the desktop
				("InsertListing", None, "Insert Listing...", None, None, lambda x: self.__listing_controller.run_dialog()),
				("InsertEntry", None, "Insert BibTeX Entry...", "<Ctrl><Alt>E", None, lambda x: self.run_bibtex_entry_dialog()),
				("Build", STOCK_CONVERT, "Build LaTeX Document", None, None, None),
				("SelectLanguage", STOCK_EDIT, "Select Language Features", None, None, None)
		])
		
		self.__ui_manager.insert_action_group(self.__action_group, -1)
		self.__ui_id = self.__ui_manager.add_ui_from_file(Installation().path + "/ui.xml")
		
		
		# add editing features selection
		features_submenu = self.__ui_manager.get_widget("/MenuBar/ExtraMenu_1/LatexMenu/SelectLanguageMenu").get_submenu()
		
		self.__listen_to_language_toggled = True
		
		self.__item_features_none = RadioMenuItem(None, "None")
		self.__item_features_none.connect("toggled", self.__language_toggled)
		self.__item_features_none.show()
		
		self.__item_features_latex = RadioMenuItem(self.__item_features_none, "LaTeX")
		self.__item_features_latex.connect("toggled", self.__language_toggled)
		self.__item_features_latex.show()
		
		self.__item_features_bibtex = RadioMenuItem(self.__item_features_none, "BibTeX")
		self.__item_features_bibtex.connect("toggled", self.__language_toggled)
		self.__item_features_bibtex.show()
		
		features_submenu.insert(self.__item_features_none, -1)
		features_submenu.insert(self.__item_features_latex, -1)
		features_submenu.insert(self.__item_features_bibtex, -1)
		
		debug(self, "__insert_widgets", "adding menu entries done, merging profiles")
		
		# insert profiles

		self.__merge_profiles()
		
		self.__init_toolbar()
		
		debug(self, "__insert_widgets", "done")
	
	def __comment_selection(self, uncomment=False):
		helper = self.__document_helper_manager.active_helper
		
		if helper.get_language() == LANGUAGE_LATEX:
			helper.language_helper.comment_selection(uncomment)
		else:
			error(self, "__comment_selection", "No LaTeX")
	
	def __remove_widgets(self):
		# remove panels
		self.__symbol_controller.deactivate()
		self.__snippet_controller.deactivate()
		self.__build_controller.deactivate()
		
		self.__window.get_side_panel().remove_item(self.__outline_view_container)
		
		# remove menu items
		self.__unmerge_profiles()
		
		self.__ui_manager.remove_ui(self.__ui_id)
		self.__ui_manager.remove_action_group(self.__action_group)
		self.__ui_manager.ensure_update()
		
		# remove toolbar
		self.main_box.remove(self.__toolbar)
	
	def __language_toggled(self, radio_menu_item):
		"""
		"""
		if not self.__listen_to_language_toggled:
			return
		
		if self.__item_features_none.get_active():
			f = LANGUAGE_NONE
		elif self.__item_features_latex.get_active():
			f = LANGUAGE_LATEX
		elif self.__item_features_bibtex.get_active():
			f = LANGUAGE_BIBTEX
			
		self.__document_helper_manager.active_helper.set_language(f)
	
	def __init_toolbar(self):
		"""
		Create and hook the toolbar
		"""
		
		debug(self, "__init_toolbar", "creating toolbar")
		
		# create
		self.__toolbar = LaTeXToolbar(self)
		
		# hook
		self.main_box = self.__window.get_children()[0]
		self.main_box.pack_start(self.__toolbar, False)
		self.main_box.reorder_child(self.__toolbar, 2)
		
		# show if set
		if self.settings["ToolbarMode"] == "on":
			self.__toolbar.show_all()
		
		debug(self, "__insert_widgets", "creating toolbar done")
		
	def __merge_profiles(self):
		"""
		Merge menu entries for the build profiles
		"""
		
		# create XML definition

		ui_top = """<ui>
				<menubar name="MenuBar">
					<placeholder name="ExtraMenu_1">
						<menu name="LatexMenu" action="Latex">
							<menu name="BuildMenu" action="Build">"""
					
		ui_bottom = """</menu>
							</menu>
						</placeholder>
					</menubar>
				</ui>"""
		
		s = ""
		actions = []
		for profile in self.__settings.list_profiles():
			s += "<menuitem name=\"%s_Menu\" action=\"%s\"/>" % (profile.name, profile.name)
			actions.append(("%s" % profile.name, None, profile.name, profile.accel, None, self.__profile_activated))
		
		ui = ui_top + s + ui_bottom

		# create actions

		self.__group_profiles = ActionGroup("LaTeXPluginBuildActions")
		self.__group_profiles.add_actions(actions)
		
		self.__ui_manager.insert_action_group(self.__group_profiles, -1)

		# merge

		self.__ui_profiles = self.__ui_manager.add_ui_from_string(ui)
	
	def __unmerge_profiles(self):
		self.__ui_manager.remove_ui(self.__ui_profiles)
		self.__ui_manager.remove_action_group(self.__group_profiles)
		self.__ui_manager.ensure_update()
	
	def __update_profiles(self):
		self.__unmerge_profiles()
		self.__merge_profiles()
	
	def __profile_activated(self, action):
		self.build(self.settings.get_profile(action.get_name()))
	
	def build(self, profile):
		self.__build_controller.build(profile)
	
	def __line_selected(self, widget, filename, line):
		"""
		A widget like the BuildView or the StructureView signaled that a source
		line has been selected
		"""
		debug(self, "__line_selected", "%s, %s : %s" % (type(widget), filename, line))
		
		try:
			helper = self.__document_helper_manager.activate_helper(filename)
			helper.jump_to_line(line)
		except KeyError:
			error(self, "__line_selected", "Failed")
	
	def run_quickstart_dialog(self):
		if not self.__dlg_quickstart:
			self.__dlg_quickstart = QuickstartDialog(self)
		
		if self.__dlg_quickstart.run() == RESPONSE_OK:
			helper = self.__document_helper_manager.active_helper
			
			helper.set_language(LANGUAGE_LATEX)
			
			helper.language_helper.insert_source(self.__dlg_quickstart.source)
			
		self.__dlg_quickstart.hide()
	
	def run_table_dialog(self):
		if not self.__dlg_table:
			self.__dlg_table = TableDialog(self.window)
		
		helper = self.__document_helper_manager.active_helper
		if helper.get_language() != LANGUAGE_LATEX:
			return
		
		latex_helper = helper.language_helper
		
		if self.__dlg_table.run(latex_helper.context) == RESPONSE_OK:
			latex_helper.insert_source(self.__dlg_table.source)
			
		self.__dlg_table.hide()
	
	def run_graphics_dialog(self):
		helper = self.__document_helper_manager.active_helper
		if helper.get_language() != LANGUAGE_LATEX:
			return
		
		filename = helper.filename
		folder = ""
		if filename:
			folder = dirname(filename)
		
		if not self.__dlg_graphics:
			self.__dlg_graphics = GraphicsDialog(self.window)
		
		if self.__dlg_graphics.run(folder) == RESPONSE_OK:
			helper.language_helper.insert_source(self.__dlg_graphics.source)
			
		self.__dlg_graphics.hide()
	
	def run_bibtex_entry_dialog(self):
		helper = self.__document_helper_manager.active_helper
		if helper.get_language() != LANGUAGE_BIBTEX:
			return
		
		bibtex_helper = helper.language_helper
		
		dlg = BibTeXEntryDialog(self.window)
		if dlg.run() == RESPONSE_OK:
			bibtex_helper.insert_source(dlg.source)
		
		dlg.hide()
	
	@property
	def listing_controller(self):
		return self.__listing_controller
	